/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2008-2010 EBM WebSourcing - http://www.petalslink.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.se.jsr181;

import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.messaging.MessagingException;
import javax.xml.namespace.QName;
import javax.xml.transform.Source;

import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.soap.SOAPBody;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFactory;
import org.apache.axiom.soap.SOAPFault;
import org.apache.axiom.soap.SOAPFaultCode;
import org.apache.axiom.soap.SOAPFaultReason;
import org.apache.axiom.soap.SOAPFaultText;
import org.apache.axiom.soap.SOAPFaultValue;
import org.apache.axis2.AxisFault;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.description.AxisOperation;
import org.apache.axis2.description.AxisService;
import org.apache.axis2.description.TransportOutDescription;
import org.apache.axis2.engine.AxisEngine;
import org.apache.axis2.util.XMLUtils;
import org.apache.axis2.wsdl.WSDLConstants;
import org.apache.commons.io.IOExceptionWithCause;
import org.ow2.petals.commons.exception.ExceptionUtil;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.api.util.MtomUtil;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.jsr181.JBIContext;
import org.ow2.petals.jsr181.JBIContextManager;
import org.ow2.petals.jsr181.PetalsMessageSender;
import org.ow2.petals.se.jsr181.axis.AxisEmptyTransportSender;
import org.ow2.petals.se.jsr181.model.JaxConfigurationHandler;
import org.ow2.petals.se.jsr181.utils.AxiomHelper;
import org.w3c.dom.Document;

/**
 * The JBIListener used to handle incoming JBI messages.
 * <p>
 * This listener invokes the linked Axis2 service.
 * </p>
 *
 * @author Christophe HAMERLING - EBM WebSourcing
 * @author Vincent Zurczak - EBM WebSourcing
 * @since 1.0
 */
public class Jsr181JBIListener extends AbstractJBIListener implements PetalsMessageSender {

    /**
     * The log prefix.
     */
    private String logHint;

    /**
     * The name space for XOP attachments.
     */
    public static final String MTOM_TNS = "http://www.w3.org/2004/08/xop/include";


    /**
     * Constructor.
     */
    public Jsr181JBIListener() {
        // nothing
    }


    /*
     * (non-Javadoc)
     * @see org.ow2.petals.component.framework.listener.AbstractJBIListener
     * #onJBIMessage(org.ow2.petals.component.framework.api.message.Exchange)
     */
    @Override
    public boolean onJBIMessage( Exchange exchange ) {

        this.logHint = "Exchange " + exchange.getExchangeId();
        Logger log = getLogger();
        if( log.isLoggable( Level.INFO ))
            log.info( this.logHint + " is started to be processed." );

        // Provider role only
        if( ! exchange.isProviderRole()) {
            String msg = "The provider role is the only role being supported in the Petals-SE-Jsr181 component.";
            if( log.isLoggable( Level.INFO ))
                log.info( this.logHint + " encountered a problem. " + msg );

            this.handleError( msg, exchange );
        }

        // Go on in the processing
        else {
            ClassLoader savedClassLoader = Thread.currentThread().getContextClassLoader();
            JBIContextManager.setJBIContext( new JBIContext( exchange, this ));
            try {
                // Get the service
                String edptName = exchange.getEndpointName();
                JaxConfigurationHandler jaxHandler = ((Jsr181Se) this.component).getJaxConfigurationHandler( edptName );
                AxisService axisService = jaxHandler.getAxisService();

                // Check if the operation is compatible
                if( ! checkOperation( exchange, axisService )) {
                    String msg = this.logHint + ": cannot find the operation " + exchange.getOperation() + " in the Axis service.";
                    this.handleError( msg, exchange );
                }

                // Check if the MEP is supported by this component
                else if( exchange.isInOptionalOutPattern()) {
                    String msg = this.logHint + ": the MEP InOptionalOut is not supported by the JSR-181 component.";
                    this.handleError( msg, exchange );
                }

                // Check if the MEP is compatible with the Axis2 service
                else if( ! checkMEP( exchange, axisService )) {
                    String msg = this.logHint + ": the MEP " + exchange.getPattern() + " is not compatible with the operation MEP.";
                    this.handleError( msg, exchange );
                }

                // Delegate the message to the Axis2 service
                else {
                    // Prepare the input
                    MessageContext inMessageContext = ((Jsr181Se) this.component).getAxisContext().createMessageContext();
                    fillInEnvelope(inMessageContext, axisService, exchange);

                    // Need to switch the class loader for the JAXBMarshaller
                    Thread.currentThread().setContextClassLoader( axisService.getClassLoader());
                    AxisEngine.receive( inMessageContext );

                    // InOut => prepare the response
                    if( exchange.isInOutPattern()) {
                        MessageContext outMessageContext = inMessageContext.getOperationContext().getMessageContext(
                                org.apache.axis2.wsdl.WSDLConstants.MESSAGE_LABEL_OUT_VALUE );

                        outMessageContext.setProperty(
                                org.apache.axis2.jaxws.spi.Constants.CACHE_CLASSLOADER,
                                axisService.getClassLoader());

                        this.fillOutJBIMessage( outMessageContext, exchange );
                    }

                    // Log the successful processing
                    if( log.isLoggable( Level.INFO ))
                        log.info( this.logHint + " was successfully processed." );
                }

            } catch( AxisFault axisFault ) {
                this.handleFault( axisFault, exchange );

            } catch( MessagingException me ) {
                this.handleFault( me, exchange );

            } catch (IOExceptionWithCause e) {
                this.handleFault(e, exchange);

            } catch( Exception e ) {
                this.handleFault( e, exchange );

            } finally {
                Thread.currentThread().setContextClassLoader( savedClassLoader );
                JBIContextManager.setJBIContext( null );
            }
        }

        return true;
    }


    /**
     * Checks if the MEP of the JBI message exchange and the Axis Operation are the same.
     * @param exchange the exchange
     * @param axisService the Axis service
     * @return true if the MEP are identical, false otherwise
     */
    protected boolean checkMEP( Exchange exchange, AxisService axisService ) {

        AxisOperation operation = axisService.getOperation( exchange.getOperation());

        Logger logger = getLogger();
        if( logger.isLoggable(Level.FINE )) {
            logger.fine( this.logHint + ": MEP for Axis Operation is " + operation.getMessageExchangePattern() + "." );
            logger.fine( this.logHint + ": MEP for JBI message is " + exchange.getPattern() + "." );
        }

        int mepConstant = operation.getAxisSpecificMEPConstant();
        return( exchange.isInOnlyPattern() && mepConstant == WSDLConstants.MEP_CONSTANT_IN_ONLY )
        || ( exchange.isInOptionalOutPattern() && mepConstant == WSDLConstants.MEP_CONSTANT_IN_OPTIONAL_OUT )
        || ( exchange.isInOutPattern() && mepConstant == WSDLConstants.MEP_CONSTANT_IN_OUT )
        || ( exchange.isRobustInOnlyPattern() && mepConstant == WSDLConstants.MEP_CONSTANT_ROBUST_IN_ONLY );
    }


    /**
     * Checks if an operation with the JBI exchange operation name exists in the Axis Service.
     * @param exchange the exchange
     * @param axisService the target Axis service
     * @return true if the exchange's operation exists in the Axis service, false otherwise
     */
    protected boolean checkOperation( Exchange exchange, AxisService axisService ) {

        AxisOperation op = axisService.getOperation( exchange.getOperation());
        return op != null;
    }


    /**
     * Creates the exchange response from a message context (SOAP envelope).
     * @param messageContext the message context, which contains the SOAP envelope
     * @param exchange the exchange on which to set the response
     * @throws Exception if something went wrong
     */
    private void fillOutJBIMessage(MessageContext messageContext, Exchange exchange)
    throws Exception {

        SOAPEnvelope response = messageContext.getEnvelope();
        Source source = AxiomHelper.createSource(response);
        if( source != null )
            exchange.setOutMessageContent(source);
    }

    /**
     * Creates a SOAP message from an exchange.
     *
     * @param messageContext
     *            the message context into which the SOAP envelop must be set
     * @param axisService
     *            the Axis2 service
     * @param exchange
     *            the exchange acting as the source
     * @throws AxisFault
     *             if an error occurs during the transformation
     * @throws MessagingException
     *             if attachments could not be retrieved
     * @throws IOExceptionWithCause
     *             if the exchange pay load could not be copied into Axis2
     *             message
     */
    private void fillInEnvelope(MessageContext messageContext, AxisService axisService, Exchange exchange)
            throws AxisFault, MessagingException, IOExceptionWithCause {

        // Prepare the message context
        messageContext.setServerSide(true);
        messageContext.setAxisService(axisService);
        TransportOutDescription out = new TransportOutDescription("JBI");
        out.setSender(new AxisEmptyTransportSender());
        messageContext.setTransportOut(out);

        // Map JBI operations to SOAP actions
        messageContext.setSoapAction(exchange.getOperationName());

        // Create a SOAP factory for the next steps
        // TODO: get the right factory
        SOAPFactory soapFactory = OMAbstractFactory.getSOAP11Factory();

        // Get the exchange payload
        Document exchangePayload = exchange.getInMessageContentAsDocument();
        if( exchangePayload == null
                || exchangePayload.getDocumentElement() == null ) {

            SOAPFault fault = soapFactory.createSOAPFault();

            SOAPFaultCode soapFaultCode = soapFactory.createSOAPFaultCode();
            SOAPFaultValue soapFaultValue = soapFactory.createSOAPFaultValue(soapFaultCode);
            soapFaultValue.setText(new QName("http://schemas.xmlsoap.org/soap/envelope/", "Client"));
            fault.setCode(soapFaultCode);

            SOAPFaultReason soapFaultReason = soapFactory.createSOAPFaultReason();
            SOAPFaultText soapFaultText = soapFactory.createSOAPFaultText(soapFaultReason);
            soapFaultText.setText("The input document cannot be null.");
            fault.setReason(soapFaultReason);

            throw new AxisFault(fault);
        }

        // Handle SwA attachments
        // A different processing is made for MTOM attachments.
        // Each attachment is passed as both SwA and MTOM.
        // It is assumed Axis2 guesses which one is the right one.
        Set<DataHandler> dataHandlers = exchange.getInMessageAttachments();
        for (DataHandler dataHandler : dataHandlers)
            messageContext.addAttachment(dataHandler.getName(), dataHandler);

        // Handle MTOM attachments
        Set<String> contentIds = MtomUtil.extractAttachmentsIdFromPayload(exchangePayload
                .getDocumentElement());

        for (String contentId : contentIds) {
            DataHandler dataHandler = null;
            for (String attName : exchange.getInMessageAttachmentNames()) {
                if (MtomUtil.compare(contentId, attName)) {
                    dataHandler = exchange.getInMessageAttachment(attName);
                    break;
                }
            }

            // FIXME: update the log according to the logging rules
            if (dataHandler != null)
                messageContext.addAttachment(contentId, dataHandler);
            else
                getLogger().warning("No data handler was found for the attachment " + contentId);
        }

        // Copy the pay load
        SOAPEnvelope envelope = soapFactory.createSOAPEnvelope();
        OMElement bodyContent;
        try {
            bodyContent = XMLUtils.toOM(exchangePayload.getDocumentElement(), false);
        } catch (Exception e) {
            throw new IOExceptionWithCause(e);
        }

        SOAPBody body = soapFactory.createSOAPBody(envelope);
        body.addChild(bodyContent);

        envelope.addChild(body);
        messageContext.setEnvelope(envelope);
    }


    /**
     * Handles faults.
     * @param exception an exception
     * @param exchange the exchange whose processing threw the exception
     */
    private void handleFault( Exception exception, Exchange exchange ) {

        String msg = exception.getMessage() != null ? exception.getMessage() : "The JSR-181 component generated an exception (handleFault).";
        Logger log = getLogger();
        if( log.isLoggable( Level.INFO ))
            log.info( this.logHint + " encountered a problem. " + msg );

        if( exchange.isInOnlyPattern()) {
            exchange.setError(new MessagingException(msg, exception));
        } else {
            try {
                exchange.setFault( exception );

            } catch( MessagingException e ) {
                handleError( "The JSR-181 component service generated an exception and cannot manage it.", exchange );
            }
        }
    }


    /**
     * Handles errors.
     * @param message a message
     * @param exchange the exchange whose processing resulted in an error
     */
    private void handleError( String message, Exchange exchange ) {
        this.handleError(new MessagingException(message), exchange);
    }


    /**
     * Handles errors.
     * @param exception an exception
     * @param exchange the exchange whose processing threw the exception
     */
    private void handleError( Exception exception, Exchange exchange ) {

        String msg = exception.getMessage() != null ? exception.getMessage() : "The JSR-181 component generated an exception (handleError).";
        Logger log = getLogger();
        if( log.isLoggable( Level.INFO ))
            log.info( this.logHint + " encountered a problem. " + msg );

        String ename = exception.getClass().getName();
        if (exception instanceof MessagingException || ename.startsWith("java.")
                || ename.startsWith("org.w3c.")
                || ename.startsWith("org.xml.")) {
            exchange.setError(exception);

        } else {
            MessagingException msgException = new MessagingException(ExceptionUtil
                    .getExtendedMessage(exception));
            exchange.setError(msgException);
        }
    }
}
