/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2010 EBM WebSourcing - http://www.petalslink.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

package org.ow2.petals.se.jsr181.model;

import java.net.URL;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.axis2.AxisFault;
import org.apache.axis2.classloader.JarFileClassLoader;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.description.AxisService;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.util.ClassLoaderUtil;
import org.ow2.petals.se.jsr181.axis.Axis2Deployer;
import org.ow2.petals.se.jsr181.utils.Jsr181ClassLoaderUtils;

/**
 * A handler for SU elements.
 * FIXME: do we need a separate class loader for every provides?
 * (Each class loader loads the same JARs and thus increases the risk of having unclosed JARs).
 * 
 * @author Vincent Zurczak - EBM WebSourcing
 */
public class JaxConfigurationHandler {

	/**
	 * A JAX configuration (which contains data to create services).
	 */
	private final JaxConfiguration jaxConfiguration;

	/**
	 * The class loader in which the service runs.
	 */
	private JarFileClassLoader classLoader;

	/**
	 * The Axis service.
	 */
	private AxisService axisService;

	/**
	 * An object that will be used as a lock in the management of the instance's state.
	 */
	private final Object lock = new Object();



	/**
	 * Constructor.
	 * @param jaxConfiguration a JAX configuration
	 */
	public JaxConfigurationHandler( JaxConfiguration jaxConfiguration ) {
		this.jaxConfiguration = jaxConfiguration;
	}


	/**
	 * @return the axisService
	 */
	public AxisService getAxisService() {
		return this.axisService;
	}


	/**
	 * @return the end-point name associated with this handler
	 * @see org.ow2.petals.se.jsr181.model.JaxConfiguration#getEndpointName()
	 */
	public String getEndpointName() {
		return this.jaxConfiguration.getEndpointName();
	}


	/**
	 * Creates the Axis service and its class loader.
	 * <p>
	 * This method must be called when the associated SU is stared.
	 * </p>
	 * 
	 * @param axis2Deployer the Axis2 deployer
	 * @param logger the component's logger
	 * @throws PEtALSCDKException if something went wrong
	 */
	public void start( Axis2Deployer axis2Deployer, Logger logger ) throws PEtALSCDKException {

		if( logger.isLoggable( Level.FINE )) {
			String msg = ": creating an Axis service for the " + this.jaxConfiguration.getClassName() + " class.";
			logger.fine( this.jaxConfiguration.getServiceUnitName() + msg );
		}

		// Start and stop operations on a same object are exclusive.
		synchronized( this.lock ) {

			// Create the class loader
			URL[] urls = ClassLoaderUtil.getUrls( this.jaxConfiguration.getSuRootPath());
			this.classLoader = (JarFileClassLoader) Jsr181ClassLoaderUtils.createClassLoader(
					urls,
					getClass().getClassLoader());

			ClassLoader saved = Thread.currentThread().getContextClassLoader();
			try {
				Thread.currentThread().setContextClassLoader( this.classLoader );

				// Create the Axis service
				this.axisService = axis2Deployer.deployJAXWS(
						this.jaxConfiguration.getClassName(),
						this.classLoader,
						this.jaxConfiguration.getServiceURL(),
						this.jaxConfiguration.getServiceUnitName());

			} catch( Exception e ) {
				throw new PEtALSCDKException( e );

			} finally {

				// Restore the class loader
				Thread.currentThread().setContextClassLoader( saved );

				// Check the existence of the Axis service
				if( this.axisService == null ) {
					this.classLoader.destroy();
					this.classLoader = null;
					throw new PEtALSCDKException( "An Axis service could not be created from the JAX-WS class." );
				}
			}
		}
	}


	/**
	 * Unregisters the Axis service and "close" its class loader.
	 * <p>
	 * This method must be called when the associated SU is stopped.
	 * </p>
	 * 
	 * @param axisContext the Axis context kept by the component
	 * @throws PEtALSCDKException if Axis could not unregister the service
	 */
	public void stop( ConfigurationContext axisContext ) throws PEtALSCDKException {

		// Start and stop operations on a same object are exclusive.
		synchronized( this.lock ) {

			// Unregister the service
			try {
				axisContext.getAxisConfiguration().removeServiceGroup( this.axisService.getName());
				this.axisService.getAxisConfiguration().removeService( this.axisService.getName());

			} catch( AxisFault e ) {
				throw new PEtALSCDKException( e );

			} finally {

				// Bug #314758: close the class loader.
				// (on Windows, the class loader locks the JARs and the un-installation fails)
				this.axisService = null;
				this.classLoader.destroy();
				this.classLoader = null;
			}
		}
	}
}
